/*
 * sbull.c 
 *
 * Simple block device driver
 *
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h> 
#include <linux/kernel.h> /* printk() */
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <asm/system.h>

#define MAJOR_NR sbull_major 	/* force definitions on in blk.h */
static int sbull_major; 			/* declare before blk.h */

#define DEVICE_NR(device) MINOR(device)
#define DEVICE_NAME "sbull"
#define DEVICE_REQUEST sbull_request

#define RAM_BASE 0xfe8f0000

#include <linux/blk.h>
#include <linux/blkdev.h>
#include "sbull.h"

int sbull_devs;
int sbull_rahead;
int sbull_size;
int sbull_blksize;
int sbull_hardsect;

Sbull_Dev *sbull_devices = NULL;
int *sbull_blksizes = NULL;
int *sbull_sizes = NULL;
int *sbull_hardsects = NULL;

/* function prototypes */
static int sbull_open (struct inode *inode, struct file *filp);
int sbull_release (struct inode *inode, struct file *filp);
static int sbull_transfer (Sbull_Dev *device, const struct request *req);
void sbull_request (request_queue_t *q);
int init_module (void);
void cleanup_module (void);

MODULE_LICENSE ("GPL");

/* open module */
static int sbull_open (struct inode *inode, struct file *filp)
{
	printk (KERN_INFO"sbull_opened\n");

	/* to prevent compiler warning */
	inode=inode;
	filp=filp;

	return 0;
}

/* release module */
int sbull_release (struct inode *inode, struct file *filp)
{
	printk (KERN_INFO"sbull_released\n");

	/* to prevent compiler warning */
	inode=inode;
	filp=filp;

	return 0;
}

struct block_device_operations sbull_bdops = {
	open:	sbull_open,
	release:	sbull_release,
};

/* data transfer */
static int sbull_transfer (Sbull_Dev *device, const struct request *req)
{
	int size;
	unsigned char *ptr;

	ptr = device->data + req->sector * sbull_hardsect;
	size = req->current_nr_sectors * sbull_hardsect;

	/* Check overflow */
	if (ptr + size > device->data + sbull_blksize*sbull_size) {
		static int count = 0;
		if (count++ < 5)
			printk (KERN_INFO "sbull: request past end of device\n");
		return 0;
	}

	/* Data transfer */
	switch (req->cmd) {
		case READ:
			printk (KERN_INFO "READ called\n");
			memcpy (req->buffer, ptr, size); /* from sbull device to buffer */
			return 1;
		case WRITE:
			printk (KERN_INFO "WRITE called\n");
			memcpy (ptr, req->buffer, size); /* from buffer to sbull device */
			return 1;
		default:
			/* can't happen */
			return 0;
	}
}


/* simple request */
void sbull_request (request_queue_t *q)
{
	int status;	

	while (1) {
		INIT_REQUEST;
		printk (KERN_INFO" request %p: cmd %i sec %li (nr. %li)\n", CURRENT, CURRENT->cmd, CURRENT->sector, CURRENT->current_nr_sectors);
		spin_lock (&sbull_devices->lock);
		status = sbull_transfer (sbull_devices, CURRENT);
		spin_unlock (&sbull_devices->lock);
		end_request (status); /* end_request (1); success */
	}

	/* to prevent compiler warning */
	q=q;
}

/* init module */
int init_module (void)
{
	int i = 0;
	
	sbull_major = SBULL_MAJOR;
	sbull_devs = SBULL_DEVS;
	sbull_rahead = SBULL_RAHEAD;
	sbull_size = SBULL_SIZE;
	sbull_blksize = SBULL_BLKSIZE;
	sbull_hardsect = SBULL_HARDSECT;

	read_ahead [sbull_major] = sbull_rahead;

	/* register block device */
	if (register_blkdev (sbull_major, "sbull", &sbull_bdops))
		return -EIO;
	printk (KERN_INFO "sbull: inserted sbull_major=%d\n", sbull_major);

	/* initialize device info */
	sbull_devices = kmalloc (sbull_devs * sizeof (Sbull_Dev), GFP_KERNEL);
	if (!sbull_devices) {
		printk (KERN_INFO "sbull_devices: kmalloc failed\n");
		unregister_blkdev (sbull_major, "sbull");
		return -ENOMEM;
	}
	memset (sbull_devices, 0, sbull_devs * sizeof (Sbull_Dev));
	for (i=0; i<sbull_devs; i++) {
		/* device data */
		sbull_devices[i].size = 1024 * sbull_size;
	}

	sbull_sizes = kmalloc (sbull_devs * sizeof (int), GFP_KERNEL);
	if (!sbull_sizes) {
		printk (KERN_INFO "sbull_sizes: kmalloc failed\n");
		unregister_blkdev (sbull_major, "sbull");
		return -ENOMEM;
	}
	for (i=0; i<sbull_devs; i++)
		sbull_sizes[i] = sbull_size;
	blk_size [sbull_major] = sbull_sizes;

	sbull_blksizes = kmalloc (sbull_devs * sizeof (int), GFP_KERNEL);
	if (!sbull_blksizes) {
		printk (KERN_INFO "sbull_blksizes: kmalloc failed\n");
		kfree (sbull_sizes);
		kfree (blk_size [sbull_major]);
		unregister_blkdev (sbull_major, "sbull");
		return -ENOMEM;
	}
	for (i=0; i<sbull_devs; i++)
		sbull_blksizes[i] = sbull_blksize;
	blksize_size [sbull_major] = sbull_blksizes;
	
	sbull_hardsects = kmalloc (sbull_devs * sizeof (int), GFP_KERNEL);
	if (!sbull_hardsects) {
		printk (KERN_INFO "sbull_hardsects: kmalloc failed\n");
		kfree (sbull_sizes);
		kfree (blk_size [sbull_major]);
		kfree (sbull_blksizes);
		kfree (blksize_size [sbull_major]);
		unregister_blkdev (sbull_major, "sbull");
		return -ENOMEM;
	}
	for (i=0; i<sbull_devs; i++)
		sbull_hardsects[i] = sbull_hardsect;
	hardsect_size [sbull_major] = sbull_hardsects;

	/* setup queue */
	blk_init_queue (BLK_DEFAULT_QUEUE(sbull_major), sbull_request);

	/* for ramdisk */
	sbull_devices->data = vmalloc (sbull_devices->size);
	if (!sbull_devices->data)
		return -ENOMEM;

	printk (KERN_INFO "sbull_devices->data address = %p\n",sbull_devices->data);
	printk (KERN_INFO "sbull: initialized values\n");

	return 0;
}

/* cleanup module */
void cleanup_module (void)
{
	blk_cleanup_queue (BLK_DEFAULT_QUEUE (sbull_major));

	/* clear global arrays */
	blk_size [sbull_major] = NULL;
	blksize_size [sbull_major] = NULL;
	hardsect_size [sbull_major] = NULL;

	kfree (blk_size [sbull_major]);
	kfree (blksize_size [sbull_major]);
	kfree (hardsect_size [sbull_major]);

	/* clear allocated memory */
	if (sbull_sizes) kfree (sbull_sizes);
	if (sbull_blksizes) kfree (sbull_blksizes);
	if (sbull_hardsects) kfree (sbull_hardsects);

	kfree (sbull_devices);

	unregister_blkdev (sbull_major, "sbull");
	printk (KERN_INFO "sbull cleaned up :)\n");

}
