/*
 * snull.c 
 *
 * Simple Linux loopback network driver
 *
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/slab.h> 	/* kmalloc() */
#include <linux/types.h> 	/* size_t */

#include <linux/in.h>
#include <linux/netdevice.h>		/* struct device and headers */
#include <linux/etherdevice.h>	/* eth_type_trans */
#include <linux/ip.h>						/* struct iphdr */
#include <linux/tcp.h>					/* struct tcphdr */
#include <linux/skbuff.h>

#include <net/checksum.h> 

#include "snull.h" 

/* function prototypes */
int snull_open (struct net_device *dev);
int snull_release (struct net_device *dev);
int snull_config (struct net_device *dev, struct ifmap *map);
void snull_rx (struct net_device *dev, int len, unsigned char *buf);

void snull_interrupt (int irq, void *dev_id, struct pt_regs *regs);
void snull_hw_tx (char *buf, int len, struct net_device *dev);
int snull_tx (struct sk_buff *skb, struct net_device *dev);
void snull_tx_timeout (struct net_device *dev);

int snull_ioctl (struct net_device *dev, struct ifreq *rq, int cmd);
int snull_rebuild_header (struct sk_buff *skb);
int snull_header (struct sk_buff *skb, struct net_device *dev, unsigned short type, void *daddr, void *saddr, unsigned int len);
struct net_device_stats *snull_stats (struct net_device *dev);

int snull_init (struct net_device *dev);
int init_module (void);
void cleanup_module (void);

/* global variables */
static int lockup = 0;
static int timeout = SNULL_TIMEOUT;

struct snull_priv {
	struct net_device_stats stats;
	int status;
	int rx_packetlen;
	u8 *rx_packetdata;
	int tx_packetlen;
	u8 *tx_packetdata;
	struct sk_buff *skb;
	spinlock_t lock;
};

struct net_device snull_devs[2] = {
	{ init: snull_init, },
	{	init: snull_init, }
};

/* functions */

/* open call */
int snull_open (struct net_device *dev)
{
	printk (KERN_INFO "inside open...\n");
	memcpy (dev->dev_addr, "\0SNUL0", ETH_ALEN);
	dev->dev_addr[ETH_ALEN-1] += (dev - snull_devs);

	netif_start_queue (dev);
	return 0;
}

/* release call */
int snull_release (struct net_device *dev)
{
	printk (KERN_INFO "inside release...\n");
	netif_stop_queue (dev);
	return 0;
}

/* configuration changes, passed by ifconfig */
int snull_config (struct net_device *dev, struct ifmap *map)
{
	if (dev->flags & IFF_UP)
		return -EBUSY;

	if (map->base_addr != dev->base_addr) {
		printk (KERN_INFO "snull: can't change I/O address");
		return -EOPNOTSUPP;
	}

	if (map->irq != dev->irq) 
		dev->irq = map->irq;

	return 0;
}

/* receive packet */
void snull_rx (struct net_device *dev, int len, unsigned char *buf)
{
	struct sk_buff *skb;
	struct snull_priv *priv = (struct snull_priv *) dev->priv;

	printk (KERN_INFO "inside snull_rx\n");
	skb = dev_alloc_skb (len+2);
	if (!skb) {
		printk (KERN_INFO "snull rx: low on memory - packet dropped\n");
		priv->stats.rx_dropped++;
		return;
	}
	skb_reserve (skb, 2); /* align IP on 16B boundary */
	memcpy (skb_put (skb, len), buf, len);

	/* write metadata and then pass to the receive level */
	skb->dev = dev;
	skb->protocol = eth_type_trans (skb, dev);
	skb->ip_summed = CHECKSUM_UNNECESSARY;
	priv->stats.rx_packets++;
	priv->stats.rx_bytes += len;
	netif_rx (skb);
	return;
}

/* snull interrupt */
void snull_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
	int statusword;
	struct snull_priv *priv;

	struct net_device *dev = (struct net_device *) dev_id;

	irq=irq; regs=regs;

	if (!dev /* paranoid */ ) return;

	/* lock the device */
	priv = (struct snull_priv *) dev->priv;
	spin_lock (&priv->lock);

	/* retrieve statusword: real netdevices use I/O instructions */
	statusword = priv->status;
	if (statusword & SNULL_RX_INTR) {
		/* send it to snull_rx for handling */
		printk (KERN_INFO "snull_interrupt: SNUL_RX_INTR\n");
		snull_rx (dev, priv->rx_packetlen, priv->rx_packetdata);
	}
	if (statusword & SNULL_TX_INTR) {
		printk (KERN_INFO "snull_interrupt: SNUL_TX_INTR\n");
		priv->stats.tx_packets++;
		priv->stats.tx_bytes += priv->tx_packetlen;
		dev_kfree_skb (priv->skb);
	}

	/* unlock the device */
	spin_unlock (&priv->lock);
	return;
}


/* transmit packet (low level interface) */
void snull_hw_tx (char *buf, int len, struct net_device *dev)
{
	struct iphdr *ih;
	struct net_device *dest;
	struct snull_priv *priv;
	u32 *saddr, *daddr;

	if (len < (sizeof (struct ethhdr) + sizeof (struct iphdr)) ) {
		printk (KERN_INFO "snull: Hmm. packet too short (%i octets)\n", len);
		return;
	}

	printk (KERN_INFO "inside snull_hw_tx\n");

	ih = (struct iphdr *) (buf + sizeof (struct ethhdr));
	saddr = &ih->saddr;
	daddr = &ih->daddr;

	((u8 *) saddr) [2] ^=1;
	((u8 *) daddr) [2] ^=1;

	ih->check = 0;
	ih->check = ip_fast_csum ((unsigned char *)ih, ih->ihl);

	if (dev == snull_devs)
		printk (KERN_INFO "%08x:%05i --> %08x:%05i\n", ntohl (ih->saddr), ntohs (((struct tcphdr *)(ih+1))->source), ntohl (ih->daddr), ntohs (((struct tcphdr *)(ih+1))->dest));
	else
		printk (KERN_INFO "%08x:%05i <-- %08x:%05i\n", ntohl (ih->daddr), ntohs (((struct tcphdr *)(ih+1))->dest), ntohl (ih->saddr), ntohs (((struct tcphdr *)(ih+1))->source));

	/* 	packet now ready for transmission; first simulate a receive interrupt
   *  on the twin device, then a transmission-done on the transmitting device 
   */

	dest = snull_devs + (dev==snull_devs ? 1 : 0);
	priv = (struct snull_priv *) dest -> priv;
	priv->status = SNULL_RX_INTR;
	priv->rx_packetlen = len;
	priv->rx_packetdata = buf;
	snull_interrupt (0, dest, NULL);

	priv=(struct snull_priv *) dev->priv;
	priv->status = SNULL_TX_INTR;
	priv->tx_packetlen = len;
	priv->tx_packetdata = buf;

	if (lockup && ((priv->stats.tx_packets + 1) % lockup) == 0) {
		/* simulate a dropped transmit interrupt */
		netif_stop_queue (dev);
		printk (KERN_INFO "simulate lockup at %ld, txp %ld\n", jiffies, (unsigned long) priv->stats.tx_packets);
	}
	else
		snull_interrupt (0, dev, NULL);
}


/* transmit a packet */
int snull_tx (struct sk_buff *skb, struct net_device *dev)
{
	int len;
	char *data;
	struct snull_priv *priv = (struct snull_priv *) dev->priv;

	printk (KERN_INFO "inside snull_tx\n");
	len = skb->len < ETH_ZLEN ? ETH_ZLEN : skb->len;
	data = skb->data;
	dev->trans_start = jiffies;

	priv->skb = skb;

	snull_hw_tx (data, len, dev);
	
	return 0;
}

void snull_tx_timeout (struct net_device *dev)
{
	struct snull_priv *priv = (struct snull_priv *) dev->priv;
	
	printk (KERN_INFO "Transmit timout at %ld, latency %ld", jiffies, jiffies - dev->trans_start);
	priv->status = SNULL_TX_INTR;
	snull_interrupt (0, dev, NULL);
	priv->stats.tx_errors++;
	netif_wake_queue (dev);
	return;
}


/* ioctl */
int snull_ioctl (struct net_device *dev, struct ifreq *rq, int cmd)
{
	dev=dev; rq=rq; cmd=cmd;
	printk (KERN_INFO "inside ioctl\n");
	return 0;
}

/* rebuild_header */
int snull_rebuild_header (struct sk_buff *skb)
{
	struct ethhdr *eth = (struct ethhdr *) skb->data;
	struct net_device *dev = skb->dev;

	memcpy (eth->h_source, dev->dev_addr, dev->addr_len);
	memcpy (eth->h_dest, dev->dev_addr, dev->addr_len);
	eth->h_dest[ETH_ALEN-1] ^= 0x01; /*destination is us, xor 1 */
	return 0;
}

/* snull header */
int snull_header (struct sk_buff *skb, struct net_device *dev, unsigned short type, void *daddr, void *saddr, unsigned int len)
{
	struct ethhdr *eth = (struct ethhdr *) skb_push (skb, ETH_HLEN);

	len=len;

	eth->h_proto = htons (type);
	memcpy (eth->h_source, saddr ? saddr : dev->dev_addr, dev->addr_len);
	memcpy (eth->h_dest, daddr ? daddr : dev->dev_addr, dev->addr_len);
	eth->h_dest [ETH_ALEN-1] ^= 0x01; /* destination is us, xor 1 */
	return (dev->hard_header_len);
}

/* status function */
struct net_device_stats *snull_stats (struct net_device *dev)
{
	struct snull_priv *priv = (struct snull_priv *) dev->priv;
	return &priv->stats;
}

/* init function */
int snull_init (struct net_device *dev)
{
	ether_setup (dev);

	dev->open							=	snull_open;
	dev->stop							=	snull_release;
	dev->set_config				=	snull_config;
	dev->hard_start_xmit	=	snull_tx;
	dev->do_ioctl					=	snull_ioctl; 				
	dev->get_stats				=	snull_stats;
	dev->rebuild_header		=	snull_rebuild_header;
	dev->hard_header			=	snull_header;
	dev->tx_timeout				= snull_tx_timeout;
	dev->watchdog_timeo		= timeout;

	dev->flags							|=	IFF_NOARP;
	dev->hard_header_cache	 = NULL;
	 
	SET_MODULE_OWNER (dev);

	dev->priv = kmalloc (sizeof (struct snull_priv), GFP_KERNEL);
	if (dev->priv == NULL)
		return -ENOMEM;
	memset (dev->priv, 0, sizeof (struct snull_priv));
	spin_lock_init (& ((struct snull_priv *) dev->priv) -> lock);

	return 0;
}


MODULE_LICENSE ("GPL");

/* init module */
int init_module (void)
{
	int i, result, device_present=0;

	printk (KERN_INFO "inside init_module\n");
	strcpy (snull_devs[0].name, "sn0");
	strcpy (snull_devs[1].name, "sn1");
	
	for (i=0; i<2; i++) 
		if ( (result = register_netdev (snull_devs + i)) )
			printk (KERN_INFO "snull: error %i registering device %s\n", result, snull_devs[i].name);
		else
			device_present++;

	return device_present ? 0 : -ENODEV;
}

/* cleanup module */
void cleanup_module (void)
{
	int i;
	for (i=0; i<2; i++) {
		kfree (snull_devs[i].priv);
		unregister_netdev (snull_devs + i);
	}

	printk (KERN_INFO "cleanup_module ... done\n");
	return;
}
