/*
 * sleepy.c 
 *
 * Put read on wait and wake up on write
 *
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */
 
#include <linux/module.h>
#include <linux/config.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/fs.h>

#include <asm/uaccess.h>

MODULE_LICENSE("GPL");

/* function prototypes */
ssize_t sleepy_write (struct file *fp, const char *buf, size_t count, loff_t *f_pos); 
ssize_t sleepy_read (struct file *filp, char *buffer, size_t count, loff_t *pos); 
int sleepy_open (struct inode *inode, struct file *file); 
int sleepy_release (struct inode *inode, struct file *file); 
int init_module (void); 
void cleanup_module (void); 

wait_queue_head_t my_queue;
DECLARE_WAIT_QUEUE_HEAD (my_queue);

/* write module */
ssize_t sleepy_write (struct file *fp, const char *buf, size_t count, loff_t *f_pos) 
{
  printk (KERN_WARNING "process %i (%s) awakening the readers...\n", current->pid, current->comm);
	wake_up_interruptible (&my_queue);
	return count;

	/* to prevent compiler warnings */
	fp = fp;
	buf = buf;
	f_pos = f_pos;
}

/* read module */
ssize_t sleepy_read (struct file *filp, char *buffer, size_t count, loff_t *pos) 
{
  printk (KERN_WARNING "process %i (%s) going to sleep\n", current->pid, current->comm);
	interruptible_sleep_on (&my_queue);
	printk (KERN_WARNING " awoken %i (%s)\n", current->pid, current->comm);
	return 0;

	/* to prevent compiler warnings */
	filp = filp;
	buffer = buffer;
	pos = pos;
	count = count;
}

/* open module */
int sleepy_open (struct inode *inode, struct file *file) 
{
	printk (KERN_WARNING"\nsleepy Data generator opened\n");
	return 0;

	/* to prevent compiler warnings */
	file = file;
	inode = inode;
}

/* release module */
int sleepy_release (struct inode *inode, struct file *file) 
{
	printk (KERN_WARNING"\nsleepy Data generator released\n");
	return 0;

	/* to prevent compiler warnings */
	file = file;
	inode = inode;
}

struct file_operations sleepy_fops = {
	read:	sleepy_read,
	write: sleepy_write,
	open:	sleepy_open,
	release:	sleepy_release,
};

/* init module */
int init_module (void) 
{
	int result;
	unsigned int sleepy_major=241;
	
	printk (KERN_WARNING"\nsleepy Data generator init module done\n");
	result = register_chrdev (sleepy_major, "sleepy", &sleepy_fops);

	if (result <0) {
		printk (KERN_WARNING "sleepy: can't get major %d\n", sleepy_major);
		return result;
	}

	if (sleepy_major==0) sleepy_major = result;
	printk (KERN_WARNING"sleepy_major = %d\n",sleepy_major);
	 
	return 0;
}

/* cleanup module */
void cleanup_module (void) 
{
	int n;
	n  = unregister_chrdev (241, "sleepy");
	if (n < 0 ) {
		printk (KERN_WARNING "cleanup: cannot unregister_chrdev %d",n);
	}
	
	else {
		printk (KERN_WARNING "sleepy cleaned up :)");
	}
}
